//
#include "heap.h"
#include "pmm.h"
#include "vmm.h"

/***memory structure(virtual address)
*   0...4M...0xD0000000...0xFF000000...0xFFBFF000...0xFFC00000...4G
*   kernel           heap          stack      page_dictory      page_table
*               |------heap-----------------|
*/

static void alloc_chunk(uint32_t start, uint32_t len);
static void free_chunk(header_t *chunk);
static void split_chunk(header_t *chunk, uint32_t len);
static void glue_chunk(header_t *chunk);

uint32_t heap_max = HEAP_START;
header_t *heap_first = 0;

void init_heap()
{

}

void *kmalloc(uint32_t l)
{
    l += sizeof(header_t);
    header_t *cur_header = heap_first, *prev_header = 0;

    while(cur_header)
    {
        if(cur_header->allocated == 0 && cur_header->length>=l)
        {
            split_chunk(cur_header,l);
            cur_header->allocated = 1;
            return (void*)((uint32_t)cur_header + sizeof(header_t));
        }
        prev_header = cur_header;
        cur_header = cur_header->next;
    }

    //have no suit free space
    uint32_t chunk_start;
    if(prev_header)     //not empty list
        chunk_start = (uint32_t)prev_header + prev_header->length;
    else    // empty list
    {
        chunk_start = HEAP_START;
        heap_first = (header_t *)chunk_start;
    }

    alloc_chunk(chunk_start,l);
    cur_header = (header_t*)chunk_start;
    cur_header->prev = prev_header;
    cur_header->next = 0;
    cur_header->allocated = 1;
    cur_header->length = l;

    prev_header->next = cur_header;

    return (void*)(chunk_start + sizeof(header_t));
}

void kfree(void *p)
{
    header_t *header = (header_t*)((uint32_t)p - sizeof(header_t));
    header->allocated = 0;
    //uniform
    glue_chunk(header);
}

void alloc_chunk(uint32_t start, uint32_t len)
{
    while(start+len > heap_max)
    {
        uint32_t page = pmm_alloc_page();
        map(heap_max, page, PAGE_PRESENT | PAGE_WRITE);
        heap_max += 0x1000;
    }
}

void free_chunk(header_t *chunk)
{

    if(chunk->prev)
        chunk->prev->next = 0;
    else
        heap_first = 0;

    while((heap_max - 0x1000) >= (uint32_t)chunk)
    {
        heap_max -= 0x1000;
        uint32_t page;
        get_mapping(heap_max,&page);
        pmm_free_page(page);
        unmap(heap_max);
    }
}

void split_chunk(header_t *chunk, uint32_t len)
{
    if(chunk->length-len > sizeof(header_t))
    {
        // [chunk(len)   ,   newchunk(free)]
        header_t *newchunk = (header_t*)((uint32_t)chunk + chunk->length);
        newchunk->prev = chunk;
        newchunk->next = 0;     // the caller deal with relationship
        newchunk->allocated = 1;
        newchunk->length = chunk->length - len;

        chunk->next = newchunk;
        chunk->length = len;
    }
}

//merge free memory chunk
void glue_chunk(header_t *chunk)
{
    if(chunk->next && chunk->next->allocated == 0)
    {
        // merge to chunk
       chunk->length += chunk->next->length;
       chunk->next->next->prev = chunk;
       chunk->next = chunk->next->next;
    }

    if(chunk->prev && chunk->prev->allocated == 0)
    {
        // merge to chunk->prev
        chunk->prev->length += chunk->length;
        chunk->prev->next = chunk->next;
        chunk->next->prev = chunk->prev;
        chunk = chunk->prev;
    }

    // if the chunk is the last then remove the free chunk
    if(chunk->next == 0)
        free_chunk(chunk);
}

//kmalloc align page
void *kmalloc_a(uint32_t l)
{
    l+= sizeof(header_t);
    header_t *cur_header = heap_first, *prev_header = 0;

    while(cur_header)
    {

        if((uint32_t)cur_header % 0x1000)
        {
            return kmalloc(l);
        }

        uint32_t base = (uint32_t)cur_header >> 12;
        base += 0x1000;
        //else not align
        if(cur_header->allocated == 0 && (base - cur_header->length>=l))
        {
            //deal with header space
            if(base-(uint32_t)cur_header <= sizeof(header_t))
            {
                if(prev_header)
                {
                    //save cur_header
                    prev_header->length += base-(uint32_t)cur_header;
                    prev_header->next = cur_header->next;
                    int length = cur_header->length;

                    alloc_chunk(base, length-(base-(uint32_t)cur_header));
                    header_t *new_header = (header_t*)base;

                    new_header->next = prev_header->next;
                    new_header->prev = prev_header;
                    prev_header->next = new_header;

                    cur_header = new_header;
                    split_chunk(cur_header,l);
                    cur_header->allocated = 1;
                    return (void*)((uint32_t)cur_header + sizeof(header_t));
                }

            }
            else
            {
                cur_header->length = base - (uint32_t)cur_header;
                alloc_chunk(base, l);
                header_t *new_header = (header_t*)base;

                new_header->next = cur_header->next;
                new_header->prev = cur_header;
                cur_header->next = new_header;

                prev_header = cur_header;
                cur_header = new_header;
                split_chunk(cur_header,l);
                cur_header->allocated = 1;
                return (void*)((uint32_t)cur_header + sizeof(header_t));
            }
        }
        prev_header = cur_header;
        cur_header = cur_header->next;
    }
}
