#include "vmm.h"
#include "idt.h"
#include "pmm.h"
#include "panic.h"
#include "printk.h"

uint32_t *page_directory = (uint32_t*) PAGE_DIR_VIRTUAL_ADDR;
uint32_t *page_tables = (uint32_t*)PAGE_TABLE_VIRTUAL_ADDR;

page_directory_t *current_directory;

extern char pmm_paging_active;

void page_fault (registers_t *regs);

void init_vmm()
{
    int i;
    uint32_t cr0;

    register_interrupt_handler(14, &page_fault);
    page_directory_t *pd = (page_directory_t*)pmm_alloc_page();
    memset(pd,0,0x1000);

    // 映射页表和页目录表（线性地址）
    pd[0] = pmm_alloc_page() | PAGE_PRESENT | PAGE_WRITE | PAGE_USER;
    uint32_t *pt = (uint32_t*)(pd[0]&PAGE_MASK);
    for(i=0; i<1024; i++)
        pt[i] = i*0x1000 | PAGE_PRESENT | PAGE_WRITE | PAGE_USER ;

    // Assign the second-last table and zero it.
    pd[1022] = pmm_alloc_page () | PAGE_PRESENT | PAGE_WRITE | PAGE_USER;
    pt = (uint32_t*) (pd[1022] & PAGE_MASK);
    memset (pt, 0, 0x1000);
    // The last entry of the second-last table is the directory itself.
    pt[1023] = (uint32_t)pd | PAGE_PRESENT | PAGE_WRITE | PAGE_USER;
    // The last table loops back on the directory itself.
    pd[1023] = (uint32_t)pd | PAGE_PRESENT | PAGE_WRITE | PAGE_USER;

    switch_page_directory(pd);

    asm volatile("mov %%cr0, %0":"=r"(cr0));
    cr0|=0x80000000;
    asm volatile("mov %0,%%cr0"::"r"(cr0));

    //map stack
    uint32_t pt_idx = PAGE_DIR_IDX((PMM_STACK_ADDR>>12));
    page_directory[pt_idx] = pmm_alloc_page () | PAGE_PRESENT | PAGE_WRITE | PAGE_USER;
    memset ((uint32_t*)page_tables[pt_idx*1024], 0, 0x1000);

    pmm_paging_active = 1;


}

void switch_page_directory(page_directory_t *pd)
{
    current_directory = pd;
    asm volatile("mov %0,%%cr3"::"r"(pd));
}

char get_mapping(uint32_t va, uint32_t *pa)
{
    uint32_t virtual_page = va/0x1000;
    uint32_t pt_idx = PAGE_DIR_IDX(virtual_page);
    if(page_directory[pt_idx]==0)
        return 0;

    if(page_tables[virtual_page] != 0)
    {
        if(pa)
            *pa = page_tables[virtual_page] & PAGE_MASK;
        return 1;
    }
}

void page_fault(registers_t *regs)
{
    uint32_t cr2;
    asm volatile("mov %%cr2,%0":"=r"(cr2));
    uint32_t error = regs->err_code;
    printk("-----Errcode: %d-------\n",error);

    if(error & 0x1)
        printk("Page wasn't present.\n");
    if(error & 0x2)
        printk("Can not write the page.\n");
    if(error & 0x4)
        printk("The processor was running in user-mode when it was interrupted.\n");
    if(error & 0x8)
        printk("Reserved bits being overwritten.\n");
    if(error & 0x10)
        printk("The fault occurred during an instruction fetch.\n");
    printk("Page error occured at page: %x\n", cr2);
    printk("----------------------\n");
    panic("ISR 14");

}

void unmap(uint32_t va)
{
    uint32_t virtual_page = va /0x1000;  //page num
    page_tables[virtual_page] = 0;
}

void map(uint32_t va, uint32_t pa, uint32_t flags)
{
    uint32_t virtual_page = va/0x1000;
    uint32_t pt_idx = PAGE_DIR_IDX(virtual_page);   //page table index

    if(page_directory[pt_idx] == 0)
    {
        page_directory[pt_idx] = pmm_alloc_page() | PAGE_PRESENT | PAGE_WRITE | PAGE_USER;
        memset((uint8_t*)page_tables[pt_idx*1024],0,0x1000);
    }

    page_tables[virtual_page] = (pa & PAGE_MASK) | flags;
}
